/*******************************************************************************
***
*** Filename         : ncpscope.c
*** Purpose          : PLP/NCP/RFSV protocol analyser. Takes input from scope.
*** Author           : Matt J. Gumbley
*** Last updated     : 02/01/99
***
********************************************************************************
***
*** Modification Record
***
*******************************************************************************/

#include <sys/times.h>
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>

#include "global.h"
#include "defs.h"
#include "plp.h"
#include "ncp.h"
#include "link.h"
#include "rfsv.h"
#include "util.h"

#define MAXLEN 80 /* Max. Length of scope input (it's less than this) */
static char buffer[MAXLEN];
#define TIMELEN 13 /* Length of the time part of the scope input */

/* XXX: I know that the Finite State Machine for plp_receive could be simpler, 
 *      but this one works OK, and I'll improve it later.
 *      Ideas: have a "recording" state, switched on by 10 02, and off by 10 03
 *             with the CRC being built up as the frame is received, rather
 *             than by running through the buffer when the 10 03 has been rxed.
 */

#define WAIT_FOR_SYN 0
#define WAIT_FOR_INITIAL_10 1
#define WAIT_FOR_INITIAL_02 2
#define WAIT_FOR_ESCAPE 3
#define WAIT_FOR_CRC1 4
#define WAIT_FOR_CRC2 5

/* This holds all state necessary for a complete PLP/NCP/RFSV&LINK 
 * implementation. We have two of these in the scope, since
 * we're monitoring two separate (but communicating) protocol stacks.
 */
struct stack_state {
  struct plp_state plp;
  char starttime[TIMELEN+1];
  struct ncp_channel ncp_channels[NCP_CHANNELS];
};

#define PSION 0
#define PC 1
struct stack_state stack_states[2];



/*******************************************************************************
***
*** Function         : plp_init(p)
*** Preconditions    : p is a plp_state structure
*** Postconditions   : The PLP layer is initialised
***
*******************************************************************************/

void plp_init(struct plp_state *p)
{
  p->buffer = NULL;
  p->bufptr = NULL;
  p->state  = WAIT_FOR_SYN;
  p->in_escape = FALSE;
}


/*******************************************************************************
***
*** Function         : link_alloc()
*** Preconditions    : A new link channel has been opened.
*** Postconditions   : The link layer is initialised, and a clean link_state 
***                    struct allocated and returned
***
*******************************************************************************/

struct link_state *link_alloc(void)
{
struct link_state *r;
  r = calloc(1, sizeof(struct link_state));
  if (r) {
    /* ??? */
  }
  else {
    fprintf(stderr,"link_alloc: OOM\n");
  }
  return r;
}

/*******************************************************************************
***
*** Function         : link_free(r)
*** Preconditions    : An link channel has been closed. r is a pointer to 
***                    its state block
*** Postconditions   : The link layer is destroyed, and freed.
***
*******************************************************************************/

void link_free(struct link_state *r)
{
  free(r);
}

/*******************************************************************************
***
*** Function         : link_analyse
*** Preconditions    : s is stack context record with buffer and length set.
***                    buf points to the start of a PLP buffer, starting in 
***                    TYPE, len is the total length excluding 10 03 CRC1 CRC2
*** Frame format     : PLP-TYPE LINKTXCHAN LINKRXCHAN link-protocol-bytes
*** Purpose          : Analyses the LINK data
***
*******************************************************************************/

static void link_analyse(struct stack_state *s)
{
/*struct plp_state *p = &s->plp;
byte *buf=p->buffer;
int len=p->length;

byte *ncpbuf=buf+1;
int ncplen=len-1;
byte local_id=ncpbuf[0];
byte remote_id=ncpbuf[1];
byte counter=ncpbuf[2];
word16 command=ncpbuf[3] | (ncpbuf[4] << 8);
word16 size=ncpbuf[5] | (ncpbuf[6] << 8);
byte *linkdata=ncpbuf+7;
int linkdatalen=ncplen-7;
*/
  printf("LINK: don't know LINK protocol yet\n");

}



/*******************************************************************************
***
*** Function         : rfsv_alloc()
*** Preconditions    : A new RFSV channel has been opened.
*** Postconditions   : The RFSV layer is initialised, and a clean rfsv_state 
***                    struct allocated and returned
***
*******************************************************************************/

struct rfsv_state *rfsv_alloc(void)
{
struct rfsv_state *r;
int i;
  r = calloc(1, sizeof(struct rfsv_state));
  if (r) {
    for (i=0; i<MAX_OPEN_FILES; i++) {
      r->open_files[i].active = FALSE;
    }
    r->last_cmd = 0xffff;
  }
  else {
    fprintf(stderr, "rfsv_alloc: OOM\n");
  }
  return r;
}

/*******************************************************************************
***
*** Function         : rfsv_free(r)
*** Preconditions    : An RFSV channel has been closed. r is a pointer to 
***                    its state block
*** Postconditions   : The RFSV layer is destroyed, and freed.
***
*******************************************************************************/

void rfsv_free(struct rfsv_state *r)
{
  free(r);
}


/*******************************************************************************
***
*** Function         : rfsv_analyse
*** Preconditions    : s is stack context record with buffer and length set.
***                    buf points to the start of a PLP buffer, starting in 
***                    TYPE, len is the total length excluding 10 03 CRC1 CRC2
*** Frame format     : PLP-TYPE LINKTXCHAN LINKRXCHAN rfsv-protocol-bytes
*** Purpose          : Analyses the RFSV data
***
*******************************************************************************/

static void rfsv_analyse(struct stack_state *s)
{
struct plp_state *p = &s->plp;
byte *buf=p->buffer;
int len=p->length;

byte *ncpbuf=buf+1;
int ncplen=len-1;
byte remote_id=ncpbuf[0];
byte local_id=ncpbuf[1];
byte counter=ncpbuf[2]; /* Eh? */
word16 command=ncpbuf[3] | (ncpbuf[4] << 8);
/*word16 size=ncpbuf[5] | (ncpbuf[6] << 8);*/

byte *rfsvdata=ncpbuf+7;              /* Points to the data area of the frame */
int rfsvdatalen=ncplen-7;

struct ncp_channel *n;
struct rfsv_state *r;
struct stack_state *sremote;
struct ncp_channel *nremote;
struct rfsv_state *rremote;
char *filename, *modestr, *serstr, *accstr, *ldstr;
byte *block;
int remaining;
word16 mode;
word16 fd;
word16 numbytes;
sword16 response;
word32 position;
word16 version;
word16 status;
word32 size;
word32 amend;
word32 reserve;
word16 type;
word16 changeable;
word32 free;
char *volname, *devname;
word16 battery;
 
  printf("RFSV: Frame %d ", counter);
  n = &s->ncp_channels[local_id];
  r = (struct rfsv_state *)n->as;
  if (!r) {
    printf("No RFSV context - not connected?\n");
    return;
  }

  sremote = (s == &stack_states[PC] ? &stack_states[PSION] : &stack_states[PC]);
  nremote = &sremote->ncp_channels[remote_id];
  rremote = (struct rfsv_state *)nremote->as;

  /* Check for remaining FWRITE frames */
  if (r->in_write) {
    if (counter == 0x01) {
      /* Last frame of written data */
      r->in_write = FALSE;
    }
    /* No further analysis required */
    printf("\n");
    return;
  }

  /* Record the command for when we next get a response (this assumes that the
   * response will be seen as the next frame, so it can make use of this
   * recording!)
   */
  if (command != RFSV_RESPONSE)
    r->last_cmd = command;

  switch (command) {
    case RFSV_FOPEN:
      printf("FOPEN ");
      mode = read_word16(rfsvdata);
      filename = rfsvdata+9;
      switch (mode & P_FMT_MASK) {
        case P_FSTREAM:
          modestr = "STREAM";
          break;
        case P_FSTREAM_TEXT:
          modestr = "STREAM_TEXT";
          break;
        case P_FTEXT:
          modestr = "TEXT";
          break;
        case P_FDIR:
          modestr = "DIR";
          break;
        case P_FFORMAT:
          modestr = "FORMAT";
          break;
        case P_FDEVICE:
          modestr = "DEVICE";
          break;
        case P_FNODE:
          modestr = "NODE";
          break;
        default:
          modestr = "unknown FMT";
          /* XXX: what to do in this case? */
          break;
      }
      switch (mode & P_FSER_MASK) {
        case P_FOPEN:
          serstr = "OPEN";
          break;
        case P_FCREATE:
          serstr = "CREATE";
          break;
        case P_FREPLACE:
          serstr = "REPLACE";
          break;
        case P_FAPPEND:
          serstr = "APPEND";
          break;
        case P_FUNIQUE:
          serstr = "UNIQUE";
          break;
        default:
          serstr = "unknown FSER";
          /* XXX: what to do in this case? */
          break;
      }
      switch (mode & P_FACC_MASK) {
        case P_FUPDATE:
          accstr = "UPDATE";
          break;
        case P_FRANDOM:
          accstr = "RANDOM";
          break;
        case P_FSHARE:
          accstr = "SHARE";
          break;
        default:
          accstr = "unknown FACC";
          /* XXX: what to do in this case? */
          break;
      }
      ldstr = (mode & P_FLOWDENSITY) ? "LOWDENSITY " : " ";
      printf("'%s' (%s,%s,%s) %s(0x%04x)\n", 
             filename, modestr, serstr, accstr, ldstr, mode);
      break;

    case RFSV_FCLOSE:
      fd = read_word16(rfsvdata);
      printf("FCLOSE #%d\n", fd);
      break;

    case RFSV_FREAD:
      fd = read_word16(rfsvdata);
      numbytes = read_word16(rfsvdata+2);
      printf("FREAD #%d, %d bytes\n", fd, numbytes);
      r->in_read = TRUE;
      break;

    case RFSV_FDIRREAD:
      fd = read_word16(rfsvdata);
      printf("FDIRREAD #%d\n", fd);
      break;

    case RFSV_FDEVICEREAD:
      fd = read_word16(rfsvdata);
      printf("FDEVICEREAD #%d\n", fd);
      break;

    case RFSV_FWRITE:
      fd = read_word16(rfsvdata);
      printf("FWRITE #%d\n", fd);
      r->in_write = TRUE;
      break;

    case RFSV_FSEEK:
      fd = read_word16(rfsvdata);
      position = read_word32(rfsvdata+2);
      mode = read_word16(rfsvdata+6);
      switch (mode) {
        case P_FABS:
          modestr = "P_FABS";
          break;
        case P_FEND:
          modestr = "P_FEND";
          break;
        case P_FCUR:
          modestr = "P_FCUR";
          break;
        case P_FRSENSE:
          modestr = "P_FRSENSE";
          break;
        case P_FRSET:
          modestr = "P_FRSET";
          break;
        case P_FREWIND:
          modestr = "P_FREWIND";
          break;
        default:
          modestr = "UNKNOWN";
          break;
      }
      printf("FSEEK %d @ 0x%08x %s\n", fd, position, modestr);
      break;

    case RFSV_FFLUSH:
      printf("FFLUSH\n");
      break;

    case RFSV_FSETEOF:
      printf("FSETEOF\n");
      break;

    case RFSV_RENAME:
      printf("RENAME\n");
      break;

    case RFSV_DELETE:
      printf("DELETE\n");
      break;

    case RFSV_FINFO:
      filename = rfsvdata;
      printf("FINFO '%s'\n", filename);
      break;

    case RFSV_SFSTAT:
      printf("SFSTAT\n");
      break;

    case RFSV_PARSE:
      printf("PARSE\n");
      break;

    case RFSV_MKDIR:
      printf("MKDIR\n");
      break;

    case RFSV_OPENUNIQUE:
      printf("OPENUNIQUE\n");
      break;

    case RFSV_STATUSDEVICE:
      printf("STATUSDEVICE\n");
      break;

    case RFSV_PATHTEST:
      printf("PATHTEST\n");
      break;

    case RFSV_STATUSSYSTEM:
      printf("STATUSSYSTEM\n");
      break;

    case RFSV_CHANGEDIR:
      printf("CHANGEDIR\n");
      break;

    case RFSV_SFDATE:
      printf("SFDATE\n");
      break;

    case RFSV_RESPONSE:
      response = read_word16(rfsvdata);
      /* Hopefully this reponse is for the previous command sent by the
       * remote side - base interpretation of the response on this assumption.
       */
      switch (rremote->last_cmd) {
        case RFSV_FOPEN:
          fd = read_word16(rfsvdata+2);
          printf("FOPEN RESPONSE %d: fd %d\n", response, fd);
          break;

        case RFSV_FCLOSE:
          printf("FCLOSE RESPONSE %d\n", response);
          break;
    
        case RFSV_FREAD:
          printf("FREAD RESPONSE %d\n", response);
          break;
    
        case RFSV_FDIRREAD:
          printf("FDIRREAD RESPONSE %d: plus data\n", response);
          /* XXX: needs much more than this! */
          break;
    
        case RFSV_FDEVICEREAD:
          printf("FDEVICEREAD RESPONSE %d\n", response);
/* THIS ISN'T WORKING RIGHT 
          block = rfsvdata+2;
          remaining = rfsvdatalen-2;
          while (remaining) {
            version = read_word16(block);
            type = read_word16(block+2);
            changeable = read_word16(block+4);
            size = read_word32(block+6);
            free = read_word32(block+10);
            volname = block+14;
            battery = read_word16(block+30);
            devname = block+62;
            printf(">> v%d type 0x%04x ch? %d sz 0x%08x fr 0x%08x vol %s bt 0x%04x dev %s\n", version, type, changeable, size, free, volname, battery, devname);
            block += 78;
            remaining -= 78;
          }
*/
          break;
    
        case RFSV_FWRITE:
          printf("FWRITE RESPONSE %d\n", response);
          break;
    
        case RFSV_FSEEK:
          position = read_word32(rfsvdata+2);
          printf("FSEEK RESPONSE %d current position 0x%08x\n", 
                  response, position);
          break;

        case RFSV_FFLUSH:
          printf("FFLUSH RESPONSE %d\n", response);
          break;

        case RFSV_FSETEOF:
          printf("FSETEOF RESPONSE %d\n", response);
          break;

        case RFSV_RENAME:
          printf("RENAME RESPONSE %d\n", response);
          break;

        case RFSV_DELETE:
          printf("DELETE RESPONSE %d\n", response);
          break;

        case RFSV_FINFO:
          /* Multiple blocks of 16 bytes */
          block = rfsvdata+2;
          remaining = rfsvdatalen-2;
          printf("FINFO RESPONSE %d\n",response);
          while (remaining) {
            version = read_word16(block);
            status = read_word16(block+2);
            size = read_word32(block+4);
            amend = read_word32(block+8);
            reserve = read_word32(block+12);
            printf(">> v%d status %d size 0x%08x amend 0x%08x res 0x%08x\n", version, status, size, amend, reserve);
            block += 16;
            remaining -= 16;
          }
          break;

        case RFSV_SFSTAT:
          printf("SFSTAT RESPONSE %d\n", response);
          break;

        case RFSV_PARSE:
          printf("PARSE RESPONSE %d\n", response);
          break;

        case RFSV_MKDIR:
          printf("MKDIR RESPONSE %d\n", response);
          break;

        case RFSV_OPENUNIQUE:
          printf("OPENUNIQUE RESPONSE %d\n", response);
          break;

        case RFSV_STATUSDEVICE:
          printf("STATUSDEVICE RESPONSE %d\n", response);
          break;

        case RFSV_PATHTEST:
          printf("PATHTEST RESPONSE %d\n", response);
          break;

        case RFSV_STATUSSYSTEM:
          printf("STATUSSYSTEM RESPONSE %d\n", response);
          break;

        case RFSV_CHANGEDIR:
          printf("CHANGEDIR RESPONSE %d\n", response);
          break;

        case RFSV_SFDATE:
          printf("SFDATE RESPONSE %d\n", response);
          break;

        default:
          printf("RESPONSE to command 0x%02x: %d: \n",rremote->last_cmd, 
                 response);
      }
      break;
    default:
      printf("UNKNOWN command 0x%04x\n", command);
      break;
  }
}




/*******************************************************************************
***
*** Function         : ncp_init(n)
*** Preconditions    : n is a set of NCP_CHANNELS ncp_channel structures
*** Postconditions   : The NCP layer is initialised.
***
*******************************************************************************/

void ncp_init(struct ncp_channel *n)
{
int i;
  memset(n, 0x00, sizeof(struct ncp_channel) * NCP_CHANNELS);

  /* Initialise the channel control blocks. 0 is always in use */
  n->flags = NCP_LINK_ACTIVE;
  strcpy(n->name, "CONTROL.*");
  for (n++,i=1; i<NCP_CHANNELS; i++,n++) {
    n->flags = 0;
    n->remote_id = NCP_CHANNEL_UNASSIGNED;
    n->name[0]='\0';
  }
}


/*******************************************************************************
***
*** Function         : ncp_analyse
*** Preconditions    : s is stack context record with buffer and length set.
***                    buf points to the start of a PLP buffer, starting in 
***                    TYPE, len is the total length excluding 10 03 CRC1 CRC2
*** Purpose          : Handle a received packet that has a correct CRC
***
*******************************************************************************/

static void ncp_analyse(struct stack_state *s)
{
struct plp_state *p = &s->plp;
struct stack_state *sremote;
char *sremotename;
byte *buf=p->buffer;
/*int len=p->length;*/
byte *ncpbuf=buf+1;
/*int   ncplen=len-1;*/
byte  remote_id=ncpbuf[0];
byte  local_id=ncpbuf[1];
byte  command=ncpbuf[2];

byte *ncpdata=&ncpbuf[3];
/*int   ncpdatalen=ncplen-4;*/
struct ncp_channel *n;
word32 remotetime;
byte  assigned_id;
byte  unknown;

  n = &s->ncp_channels[local_id];
  sremote = (s == &stack_states[PC] ? &stack_states[PSION] : &stack_states[PC]);
  sremotename = (s == &stack_states[PC] ? "PSION" : "PC");

  printf("NCP: [remote %d local %d] ", remote_id, local_id);
  if (n->remote_id != NCP_CHANNEL_UNASSIGNED)
    printf("(==> remote %d) ", n->remote_id);
  if (remote_id == 0x00) {
    /* This is a control frame, so process it */
    switch (command) {
      case NCON_MSG_DATA_XOFF:
        printf("XOFF\n");
        break;

      case NCON_MSG_DATA_XON:
        printf("XON\n");
        break;

      case NCON_MSG_CONNECT_TO_SERVER:
        /* local_id is the channel we want to connect to the process.
         * (It's the channel assigned on the remote machine: we're getting a
         * frame from (say) the Psion, and ITS local_id is OUR remote...
         * Record the process so we can dispatch to the correct application
         * layer protocol analyser
         */
        printf("CONNECT_TO_SERVER %s\n", ncpdata);
        if (n->name[0] != '\0')
          printf("This channel is already connected to %s!\n", n->name);
        else {
          strncpy(n->name, ncpdata, NCP_PROCESSNAME_SIZE);
          if (strcmp(n->name, "LINK.*") == 0)
            n->as=link_alloc();
          else if (strcmp(n->name, "SYS$RFSV.*") == 0)
            n->as=rfsv_alloc();
          else
            printf("bogus! Don't support %s!\n", n->name);        
          if (!n->as)
            printf("ncp_analyse: OOM\n");
        }
        break;

      case NCON_MSG_CONNECT_RESPONSE:
        assigned_id=ncpdata[0];
        unknown=ncpdata[1];
        strncpy(n->name, sremote->ncp_channels[assigned_id].name,
                NCP_PROCESSNAME_SIZE);
        if (strcmp(n->name, "LINK.*") == 0)
          n->as=link_alloc();
        else if (strcmp(n->name, "SYS$RFSV.*") == 0)
          n->as=rfsv_alloc();
        else
          printf("bogus! Don't support %s!\n", n->name);        
        if (!n->as)
          printf("ncp_analyse: OOM\n");
        printf("CONNECT_RESPONSE assigned %s channel %d (data: 0x%02x)\n",
               sremotename, assigned_id, unknown);
        printf("CONNECT_RESPONSE this is a %s channel\n", n->name);
        break;

      case NCON_MSG_CHANNEL_CLOSED:
        printf("CHANNEL_CLOSED\n");
        break;

      case NCON_MSG_NCP_INFO:
        /* We've been sent a logon (time) frame, indicating the time at which
         * the remote NCP server started. 
         * We might get 3D 24 A6 36 in ncpdata, corresponding to 0x36A6243D
         * which is the UNIX 32-bit time value.
         */
        remotetime = ncpdata[1] | (ncpdata[2] << 8) | 
                     (ncpdata[3] << 16) | (ncpdata[4] << 24);
        printf("NCP_INFO v%d time is 0x%08x\n", *ncpdata, remotetime);
        break;

      case NCON_MSG_CHANNEL_DISCONNECT:
        printf("CHANNEL_DISCONNECT\n");
        break;

      case NCON_MSG_NCP_END:
        printf("NCP_END\n");
        break;

      default:
        printf("ncp_analyse: ** UNKNOWN TYPE 0x%02x\n",command);
        break;
    }
  }
  else {
    /* This frame is destined for a local process */
    printf("DATA for %s\n", n->name);
    /* Analyse at the application layer */
    if (strcmp(n->name, "LINK.*") == 0)
      link_analyse(s);
    else if (strcmp(n->name, "SYS$RFSV.*") == 0)
      rfsv_analyse(s);
    else
      printf("bogus! Don't support %s!\n", n->name);
  }
}


/*******************************************************************************
***
*** Function         : plp_analyse
*** Preconditions    : s is stack context record with buffer and length set.
***                    buf points to the start of a PLP buffer, starting in 
***                    TYPE, len is the total length excluding 10 03 CRC1 CRC2
*** Purpose          : Handle a received packet that has a correct CRC
***
*******************************************************************************/

static void plp_analyse(struct stack_state *s)
{
struct plp_state *p = &s->plp;
byte *buf=p->buffer;
int len=p->length;
byte type = buf[0];
byte seq = type & PLP_SEQUENCE_MASK;

  if (!buf) {
    fprintf(stderr,"WARNING: plp_analyse called on NULL buffer\n");
  }
  printf("-----------------------------------------------------------------------------\n");

  /* The data looks like... */
  hexdump(buf,len);

  /* Who sent this data? (Starttime isn't a PLP attribute, but the scope tracks
   * it anyway) 
   */
  if (s == &stack_states[PSION]) 
    printf("PSION: %s, %d bytes\n", s->starttime, len);
  else
    printf("PC:    %s, %d bytes\n", s->starttime, len);

  printf("PLP: ");

  /* Analyse at PLP level */
  switch (type & PLP_MSG_MASK) {
    case PLP_MSG_ACKNOWLEDGE:
      printf("ACKNOWLEDGE  #%d\n", seq);
      break;
      
    case PLP_MSG_DISCONNECT:
      printf("DISCONNECT   #%d\n", seq);
      break;

    case PLP_MSG_LINK_REQUEST:
      printf("LINK REQUEST #%d\n", seq);
      break;

    case PLP_MSG_DATA:
      printf("DATA         #%d\n", seq);
      /* Analyse at the NCP level */
      ncp_analyse(s);
      break;

    default:
      printf("UNKNOWN type %d, #%d\n", type & PLP_MSG_MASK, seq);
  }
  if (!buf) {
    fprintf(stderr,"WARNING: about to free NULL buffer in plp_analyse\n");
  }
  /* finished with this buffer */
  free(buf);
}



/*******************************************************************************
***
*** Function         : plp_scope_receive(ch)
*** Preconditions    : p->ch is a single character received from the scope data 
*** Postconditions   : A frame might be enqueued for the upper layer (NCP),
***                    or the character might just be built up into an incoming
***                    frame buffer.
***                    This routine returns TRUE if the character was processed
***                    and FALSE if there was an OOM error.
***
*******************************************************************************/

void plp_scope_receive(struct stack_state *s, char *scopeline)
{
struct plp_state *p = &s->plp;
#ifdef DEBUG
  fprintf(stderr,"plp_receive: rxed 0x%02x (%c) state %d\n", 
          p->ch, isprint(p->ch) ? p->ch : '?', p->state);
#endif
  if (!p->buffer) {
    p->buffer = calloc(1, PLP_FRAMESIZE+100); /* +100 just in case */
    if (!p->buffer) {
      fprintf(stderr,"plp_receive: OOM\n");
      return;
    }
    p->bufptr = p->buffer;
  }
  switch (p->state) {
    case WAIT_FOR_SYN: 
      if (p->ch != 0x16)
        return;
      p->state = WAIT_FOR_INITIAL_10;
      /* Record start time of frame */
      strncpy(s->starttime, scopeline, TIMELEN);
      return;

    case WAIT_FOR_INITIAL_10:
      if (p->ch != 0x10) {
        p->state = WAIT_FOR_SYN;
        return;
      }
      p->state = WAIT_FOR_INITIAL_02;
      return;

    case WAIT_FOR_INITIAL_02:
      if (p->ch != 0x02) {
        p->state = WAIT_FOR_SYN;
        return;
      }
      /* I used to have a separate state here, waiting for the TYPE byte, but
       * that gets escaped too, if it's an 0x10 (DISCONNECT), so it's easier to
       * just treat it (at this level) as a piece of data.
       */
      p->state = WAIT_FOR_ESCAPE;
      return;

    case WAIT_FOR_ESCAPE:
      if (p->in_escape) {
        p->in_escape = FALSE;
        /* i.e. last char was the 0x10 escape indicator */
        switch (p->ch) {
          case 0x10:
            /* A data byte of 0x10 */
            break;
          case 0x03:
            /* Finished with the data */
            p->state = WAIT_FOR_CRC1;
            /* Don't store the 0x03 */
            return;
        }
      }
      else {
        /* Could be any data byte; could be 0x10 command indicator */
        if (p->ch == 0x10) {
          /* Don't store the command indicator */
          p->in_escape = TRUE;
          return;
        }
      }
      break;

    case WAIT_FOR_CRC1:
      p->CRC = (p->ch << 0x8);
      p->state = WAIT_FOR_CRC2;
      return;

    case WAIT_FOR_CRC2:
      p->CRC |= p->ch;
#ifdef DEBUG
      fprintf(stderr,"plp_receive: rxed CRC is 0x%04x\n", p->CRC);
#endif
      /* Finished. Now is the frame ok? */
      p->length=p->bufptr-p->buffer;
#ifdef DEBUG
      fprintf(stderr,"length is %d\n", p->length);
#endif
      p->genCRC = crc(p->buffer,p->length);
#ifdef DEBUG
      fprintf(stderr,"genCRC is 0x%04x\n", p->genCRC);
#endif

      if (p->genCRC == p->CRC) {
        /* OK, good packet, decode it */
        plp_analyse(s);

        /* Give me a new buffer to fill */
        p->buffer = NULL;
      }
      else {
        printf("plp_receive: *** CRC ERROR ***\n");
        p->bufptr = p->buffer;
        /* Reuse the old buffer */
      }
      p->state = WAIT_FOR_SYN;
      return;
  }
#ifdef DEBUG
  fprintf(stderr,"building 0x%02x\n", p->ch);
#endif
  *(p->bufptr)++=(byte)(p->ch & 0xff);
  return;
}
 



int main(int argc, char *argv[])
{
int i;
struct stack_state *s;

  for (i=0; i<2; i++) {
    plp_init(&stack_states[i].plp);
    ncp_init(&stack_states[i].ncp_channels[0]);
  }

  while (fgets(buffer, MAXLEN, stdin)) {
    /* Line is of the form:
     * 0123456789012345678901234567890123456789012
     *      3.605883|16 .|    |   |   |   |   |   
     * dddddd.dddddd|hh c|hh c|CTS|DSR|DCD|RTS|DTR
     *    Event Time|  / | \  |control/status lines
     *                /     \
     *       Psion Data     PC Data
     */
    if (buffer[14] != ' ' && buffer[15] != ' ') {
      s=&stack_states[PSION];
      sscanf(buffer+14, "%02X", &s->plp.ch);
      plp_scope_receive(s, buffer);
    }
    fflush(stdout);
    if (buffer[19] != ' ' && buffer[20] != ' ') {
      s=&stack_states[PC];
      sscanf(buffer+19, "%02X", &s->plp.ch);
      plp_scope_receive(s, buffer);
    }
    fflush(stdout);
  }
  return 0;
}

